"""Unified memoLite command entrypoint.

This command wraps common workflows:
- memolite serve
- memolite init/configure/... (delegates to memolite-configure)
- memolite service ...
- memolite openclaw setup
"""

from __future__ import annotations

import argparse
import os
import subprocess
import sys
from pathlib import Path

from memlite.app.main import main as run_server
from memlite.cli import main as configure_main


def _repo_root() -> Path:
    return Path(__file__).resolve().parents[3]


def _run_script(script: str, args: list[str], env: dict[str, str] | None = None) -> int:
    script_path = _repo_root() / "scripts" / script
    if not script_path.exists():
        raise FileNotFoundError(f"script not found: {script_path}")
    cmd = [str(script_path), *args]
    merged_env = os.environ.copy()
    if env:
        merged_env.update(env)
    proc = subprocess.run(cmd, env=merged_env)
    return int(proc.returncode)


def build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(prog="memolite")
    sub = parser.add_subparsers(dest="command", required=True)

    sub.add_parser("serve", help="Run memoLite API server in foreground")

    # Keep compatibility for existing configure workflow.
    config = sub.add_parser("configure", help="Run memolite-configure subcommands")
    config.add_argument("args", nargs=argparse.REMAINDER)

    service = sub.add_parser("service", help="Manage managed memoLite service")
    service_sub = service.add_subparsers(dest="action", required=True)
    install = service_sub.add_parser("install", help="Install service definition")
    install.add_argument("--enable", action="store_true", help="Enable auto-start after install")
    service_sub.add_parser("uninstall")
    service_sub.add_parser("enable")
    service_sub.add_parser("disable")
    service_sub.add_parser("start")
    service_sub.add_parser("stop")
    service_sub.add_parser("restart")
    service_sub.add_parser("status")

    openclaw = sub.add_parser("openclaw", help="OpenClaw integration commands")
    oc_sub = openclaw.add_subparsers(dest="action", required=True)
    setup = oc_sub.add_parser("setup", help="One-shot OpenClaw+memoLite setup")
    setup.add_argument("--base-url", default="http://127.0.0.1:18731")
    setup.add_argument("--org-id", default="openclaw")
    setup.add_argument("--project-id", default="openclaw")
    setup.add_argument("--user-id", default="openclaw")
    setup.add_argument("--auto-capture", choices=["true", "false"], default="true")
    setup.add_argument("--auto-recall", choices=["true", "false"], default="true")
    setup.add_argument("--search-threshold", type=float, default=0.5)
    setup.add_argument("--top-k", type=int, default=5)

    return parser


def main(argv: list[str] | None = None) -> int:
    args = build_parser().parse_args(argv)

    if args.command == "serve":
        run_server()
        return 0

    if args.command == "configure":
        forwarded = args.args
        if forwarded and forwarded[0] == "--":
            forwarded = forwarded[1:]
        return configure_main(forwarded)

    if args.command == "service":
        cmd_args: list[str] = [args.action]
        if args.action == "install" and args.enable:
            cmd_args.append("--enable")
        return _run_script("memolite_service.sh", cmd_args)

    if args.command == "openclaw" and args.action == "setup":
        env = {
            "BASE_URL": args.base_url,
            "ORG_ID": args.org_id,
            "PROJECT_ID": args.project_id,
            "USER_ID": args.user_id,
            "AUTO_CAPTURE": args.auto_capture,
            "AUTO_RECALL": args.auto_recall,
            "SEARCH_THRESHOLD": str(args.search_threshold),
            "TOP_K": str(args.top_k),
        }
        return _run_script("setup_openclaw_memolite.sh", [], env=env)

    return 1


if __name__ == "__main__":
    sys.exit(main())



"""Configuration loading for memoLite."""

from functools import lru_cache
from pathlib import Path

from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict

DEFAULT_DATA_DIR = Path.home() / ".memolite"
DEFAULT_SQLITE_PATH = DEFAULT_DATA_DIR / "memolite.sqlite3"
DEFAULT_KUZU_PATH = DEFAULT_DATA_DIR / "kuzu"


class Settings(BaseSettings):
    """Runtime settings loaded from environment variables."""

    model_config = SettingsConfigDict(
        env_prefix="MEMLITE_",
        env_file=".env",
        env_file_encoding="utf-8",
        extra="ignore",
    )

    app_name: str = Field(default="memoLite")
    environment: str = Field(default="development")
    host: str = Field(default="127.0.0.1")
    port: int = Field(default=8080)
    log_level: str = Field(default="INFO")
    sqlite_path: Path = Field(default=DEFAULT_SQLITE_PATH)
    kuzu_path: Path = Field(default=DEFAULT_KUZU_PATH)
    sqlite_vec_extension_path: Path | None = Field(default=None)
    mcp_api_key: str | None = Field(default=None)
    semantic_search_candidate_multiplier: int = Field(default=3)
    semantic_search_max_candidates: int = Field(default=100)
    episodic_search_candidate_multiplier: int = Field(default=4)
    episodic_search_max_candidates: int = Field(default=100)

    @property
    def data_dir(self) -> Path:
        """Return the common data directory."""
        return self.sqlite_path.parent


@lru_cache(maxsize=1)
def get_settings() -> Settings:
    """Return cached application settings."""
    return Settings()


def reset_settings_cache() -> None:
    """Clear the cached settings for tests."""
    get_settings.cache_clear()



"""Application resource manager bootstrap."""

import hashlib
import math
import re
from dataclasses import dataclass, field

from memlite.app.background import BackgroundTaskRunner
from memlite.common.config import Settings
from memlite.episodic.delete import EpisodicDeleteService
from memlite.episodic.derivative_pipeline import DerivativePipeline
from memlite.episodic.search import EpisodicSearchService
from memlite.memory.config_service import MemoryConfigService
from memlite.metrics.service import MetricsService
from memlite.orchestrator.memory_orchestrator import MemoryOrchestrator
from memlite.semantic.service import SemanticService
from memlite.semantic.session_manager import SemanticSessionManager
from memlite.storage.episode_store import SqliteEpisodeStore
from memlite.storage.graph_store import KuzuGraphStore
from memlite.storage.kuzu_engine import KuzuEngineFactory
from memlite.storage.project_store import SqliteProjectStore
from memlite.storage.semantic_config_store import SqliteSemanticConfigStore
from memlite.storage.semantic_feature_store import SqliteSemanticFeatureStore
from memlite.storage.session_store import SqliteSessionStore
from memlite.storage.sqlite_engine import SqliteEngineFactory
from memlite.storage.sqlite_vec import SqliteVecIndex


_TOKEN_PATTERN = re.compile(r"[\w\-']+")


async def default_embedder(text: str) -> list[float]:
    """Return a deterministic lightweight embedding.

    This is still a local fallback embedder, but unlike the previous keyword
    bucket implementation it supports open-vocabulary text via hashed token
    projections.
    """
    dimensions = 64
    vector = [0.0] * dimensions
    lowered = text.lower()
    tokens = _TOKEN_PATTERN.findall(lowered)
    if not tokens:
        return vector

    for token in tokens:
        digest = hashlib.blake2b(token.encode("utf-8"), digest_size=8).digest()
        bucket = int.from_bytes(digest[:4], "big") % dimensions
        sign = 1.0 if (digest[4] & 1) == 0 else -1.0
        vector[bucket] += sign

    norm = math.sqrt(sum(value * value for value in vector))
    if norm == 0:
        return vector
    return [value / norm for value in vector]


@dataclass
class ResourceManager:
    """Bootstrap runtime singleton-like services."""

    settings: Settings
    metrics: MetricsService
    memory_config: MemoryConfigService
    sqlite: SqliteEngineFactory
    kuzu: KuzuEngineFactory
    project_store: SqliteProjectStore
    session_store: SqliteSessionStore
    episode_store: SqliteEpisodeStore
    semantic_config_store: SqliteSemanticConfigStore
    semantic_feature_store: SqliteSemanticFeatureStore
    derivative_index: SqliteVecIndex
    graph_store: KuzuGraphStore
    derivative_pipeline: DerivativePipeline
    episodic_search: EpisodicSearchService
    episodic_delete: EpisodicDeleteService
    semantic_service: SemanticService
    semantic_session_manager: SemanticSessionManager
    orchestrator: MemoryOrchestrator
    background_tasks: BackgroundTaskRunner
    _initialized: bool = field(default=False, init=False, repr=False)

    @classmethod
    def create(cls, settings: Settings) -> "ResourceManager":
        # Resource wiring stays in one place so API, MCP and CLI all reuse the
        # exact same runtime graph and recovery hooks.
        sqlite = SqliteEngineFactory(settings)
        kuzu = KuzuEngineFactory(settings)
        metrics = MetricsService()
        kuzu.bind_metrics(metrics)
        project_store = SqliteProjectStore(sqlite)
        session_store = SqliteSessionStore(sqlite)
        episode_store = SqliteEpisodeStore(sqlite)
        semantic_config_store = SqliteSemanticConfigStore(sqlite)
        semantic_feature_store = SqliteSemanticFeatureStore(sqlite)
        derivative_index = SqliteVecIndex(sqlite, "derivative_feature_vectors")
        derivative_index.bind_metrics(metrics)
        graph_store = KuzuGraphStore(kuzu)
        derivative_pipeline = DerivativePipeline(
            graph_store=graph_store,
            derivative_index=derivative_index,
            embedder=default_embedder,
        )
        episodic_search = EpisodicSearchService(
            episode_store=episode_store,
            graph_store=graph_store,
            derivative_index=derivative_index,
            embedder=default_embedder,
            metrics=metrics,
            candidate_multiplier=settings.episodic_search_candidate_multiplier,
            max_candidates=settings.episodic_search_max_candidates,
        )
        episodic_delete = EpisodicDeleteService(
            episode_store=episode_store,
            graph_store=graph_store,
            derivative_index=derivative_index,
        )
        semantic_service = SemanticService(
            feature_store=semantic_feature_store,
            config_store=semantic_config_store,
            embedder=default_embedder,
            default_category_resolver=lambda _set_id: [],
            candidate_multiplier=settings.semantic_search_candidate_multiplier,
            max_candidates=settings.semantic_search_max_candidates,
        )
        semantic_session_manager = SemanticSessionManager(semantic_config_store)
        orchestrator = MemoryOrchestrator(
            project_store=project_store,
            session_store=session_store,
            episode_store=episode_store,
            semantic_feature_store=semantic_feature_store,
            semantic_service=semantic_service,
            episodic_search_service=episodic_search,
            episodic_delete_service=episodic_delete,
            derivative_pipeline=derivative_pipeline,
        )
        resources = cls(
            settings=settings,
            metrics=metrics,
            memory_config=MemoryConfigService(),
            sqlite=sqlite,
            kuzu=kuzu,
            project_store=project_store,
            session_store=session_store,
            episode_store=episode_store,
            semantic_config_store=semantic_config_store,
            semantic_feature_store=semantic_feature_store,
            derivative_index=derivative_index,
            graph_store=graph_store,
            derivative_pipeline=derivative_pipeline,
            episodic_search=episodic_search,
            episodic_delete=episodic_delete,
            semantic_service=semantic_service,
            semantic_session_manager=semantic_session_manager,
            orchestrator=orchestrator,
            background_tasks=None,  # type: ignore[arg-type]
        )
        resources.background_tasks = BackgroundTaskRunner(resources=resources)
        return resources

    async def initialize(self) -> None:
        """Initialize backing stores and schemas."""
        if self._initialized:
            return
        await self.sqlite.initialize_schema()
        await self.semantic_feature_store.initialize()
        await self.derivative_index.initialize()
        await self.kuzu.initialize_schema()
        await self.background_tasks.run_startup_recovery()
        self._initialized = True

    async def close(self) -> None:
        """Close backing resources."""
        if not self._initialized:
            return
        await self.kuzu.close()
        await self.sqlite.dispose()
        self._initialized = False
